#!/usr/bin/gawk -f
#
# 	Copyright (c) 2005-2008 Peter V. Saveliev
#
# 	This file is part of rt-network utilities package.
#
# 	rt-network is free software; you can redistribute it and/or modify
# 	it under the terms of the GNU General Public License as published by
# 	the Free Software Foundation; either version 3 of the License, or
# 	(at your option) any later version.
#
# 	rt-network is distributed in the hope that it will be useful,
# 	but WITHOUT ANY WARRANTY; without even the implied warranty of
# 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# 	GNU General Public License for more details.
#
# 	You should have received a copy of the GNU General Public License
# 	along with rt-network; if not, write to the Free Software
# 	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

###
#
# fill an item array from tc output
#
# the function arguments:
#	ia	- item array (_qdisc, _class and so on)
#	type	- type to work on: ether "qdisc" or "class"
#
# local variables
#	t	- node type (qdisc, class -- htb, cbq, sfq etc.; filter -- u32, fwmark)
#	h	- handle or flowid
#	cmd	- command line to run
#
function fill_array(ia,type,  t,h,cmd) {

	cmd = "/sbin/tc -s "type" show dev "_iface
	while (cmd | getline)
	{
		# the first line -- create array item
		if ($0 ~ /^[[:alpha:]]/)
		{

			# save the unique handle (qdiscs and classes)
			if (type ~ "filter")
			{
				# filters:
				#
				# _filter[A] == count of filters for handle A
				# _filter[A,B] == count of matches for a filter
				# _filter[A,B,C] == match case No.C for a filter
				#
				# an example:
				#
				# _filter["1:5"] == 5	- there is 5 (0..4) filters for handle 1:5
				# _filter["1:5",1] == 3	- there are 3 cases for filter 1
				# _filter["1:5",0,0]	- contains a match case

				# try to extract flowid -- u32 classifier
				h = _extract("flowid",$0)
				if (h)
				{
					t = "u32"
				}
				else
				{
					# if there is no u32, look for fwmark
					h = _extract("classid",$0)
					t = "fwmark"
				}
				ia[h]++

				# if it is a fwmark filter, extract match case
				if (t ~ "fwmark")
				{
					ia[h,ia[h],"type"] = "fwmark"
					ia[h,ia[h]] = 1
					ia[h,ia[h],0] = _extract("handle",$0)
				}
				else
				{
					ia[h,ia[h],"type"] = "u32"
				}
			}
			else
			{
				# get a node type
				t = _extract(type,$0)

				h = gensub(/:$/,":0","",_extract(t,$0))
				ia[h] = "true"

				# fill in parameters for this handle
				ia[h,"type"] = t

				if ($0 ~ "parent")
				{
					p = gensub(/:$/,":0","",_extract("parent",$0))
					ia[h,"parent"] = p
					_child[p,_child[p]++] = h
				}

				if ($0 ~ "rate")
				{
					ia[h,"rate"] = _extract("rate",$0)
				}

				# all string for debug purposes
				ia[h,"all"] = $0

				# look for a root qdisc or properly fill in classes
				if (!ia[h,"parent"])
				{
					if (type ~ "qdisc")
					{
						# there may be only one root qdisc
						_qdisc["root"] = h
					}
					else
					{
						# if it is a class, attach it to a
						# root qdisc
						_child[_qdisc["root"],_child[_qdisc["root"]]++] = h
					}
				}
			}
		}
		# the next line[s] -- save statistics or filter matches
		else
		{
			if (type ~ "filter")
			{
				ia[h,ia[h],ia[h,ia[h]]++] = $0
			}
			else
			{
				ia[h,"statistics",ia[h,"statistics"]++] = $0
			}
		}
	}
	close(cmd)
}

###
#
# convert an ip address or a netmask from raw hex to dotted quad
#
# the function arguments:
#	hstr	- string to convert
#
# local variables
#	dstr	- dotted quad

function aprint(hstr,    dstr) {
	dstr = ""
	for ( i = 1; i < 8; i = i+2 ) {
		dstr = dstr"."strtonum("0x"substr(hstr,i,2))
	}
	return substr(dstr,2)
}

###
#
# dump a node
#
# the function arguments:
#	ia	- item array
#	h	- handle
#	it	- indentation
#	lc	- "last child" flag
#
# local variables
#	i	- iterator
#	k	- iterator
#	t	- match case type
#	m	- match case string
#	il	- indentation line
#
function dump(ia,h,it,lc,    i,k,t,il) {

	# create indentation line
	for (i = 0; i < it; i++)
	{
		il = il" "_indent[i]
	}

	# dump an item
	if (ia[h,"all"] ~ /^qdisc/)
	{
		print il"\b+-"h" "ia[h,"all"]
	}
	else if (ia[h,"all"] ~ /^class/)
	{
		print il"\b+-"h" class "ia[h,"type"]
	}
	# dump filters
	if (_filter[h])
	{
		print il" |\t* Filters:"
		for (i = 1 ; i <= _filter[h] ; i++)
		{
			if (_filter[h,i,"type"] ~ "fwmark")
			{
				print il" |\t\t "i". fw mark: "strtonum(_filter[h,i,0])
			}
			else if (_filter[h,i,"type"] ~ "u32")
			{
				print il" |\t\t "i". u32:"
				for (k = 0 ; k < _filter[h,i] ; k++)
				{
					t = strtonum(_extract("at",_filter[h,i,k]))
					m = _extract("match",_filter[h,i,k])
					# protocol match
					if (t == 8)
					{
						print il" |\t\t\tproto "proto[strtonum("0x"substr(gensub(/\/.*/,"","",m),1,4))]
					} else

					if ((t == 12) || (t == 16))
					{
						ipaddr = gensub(/\/.*/,"","",m)
						ntmask = gensub(/.*\//,"","",m)
						if (t == "12")
						{
							iokey = "src"
						}
						else
						{
							iokey = "dst"
						}
						print il" |\t\t\t"iokey" ipaddr "aprint(ipaddr)" netmask "aprint(ntmask)
					} else

					if (t == 20)
					{
						sport = ""
						dport = ""
						ports = gensub(/\/.*/,"","",m)
						sport = strtonum("0x"substr(ports,1,4))
						dport = strtonum("0x"substr(ports,5,4))
						if ( sport )
							print il" |\t\t\tsource port "sport
						if ( dport )
							print il" |\t\t\tdestination port "dport
					} else

					print il" |\t\t\t"_filter[h,i,k]
				}
			}
		}
	}
	
	# dump statistics
	if (ia[h,"statistics"]) print il" |\t* Statistics:"
	had_up = 0
	for (i = 0; i < ia[h,"statistics"]; i++)
	{
		line = gensub(/^[[:blank:]]*/,"","",ia[h,"statistics",i])
		if (line)
		{
			print il" |\t\t"line
			had_up = 1
		}
	}
	if (had_up == 0) {
		print il" |\t\tNone"
	}

	# add an extra empty line if there is need
	if (_child[h])
	{
		print il" |"
	}
	else
	{
		print il" `-\n"il
	}
}

###
#
# make a recursion through the queueing tree
#
# the function arguments:
#	h	- handle
#	it	- indentation level (recursion level)
#	lc	- "last child" flag
#
# local variables:
#	i	- child counter
#	l	- "last child", local copy
#
function iterate(h,it,lc,    i,l) {

	# if it is a qdisc
	if (_qdisc[h])
	{
		dump(_qdisc,h,it,lc)
	}
	# else if it is a class
	else if (_class[h])
	{
		dump(_class,h,it,lc)
	}
	# else raise an exception
	else
	{
		print "no such item collected"
	}

	# if h has a child[ren], fill in indent
	if (_child[h])
	{
		_indent[it] = "|"
	}
	else
	{
		_indent[it] = " "
	}
	# iterate childs
	for (i = 0 ; i < _child[h]; i++)
	{
		# if it is the last child, clear parent"s indent
		if (i == (_child[h] - 1))
		{
			_indent[it] = " "
			l = "true"
		}
		else
		{
			l = "false"
		}
		iterate(_child[h,i],it + 1,l)
	}
}

BEGIN {

	if (_iface ~ /.*\@.*/)
	{
		gsub(/\@.*$/,"",_iface)
	}

	profile = "/etc/protocols"
	while ( getline line < profile ) {
		if (!match(line,/^#/)) {
			split(line,apart)
			proto[strtonum(apart[2])] = apart[1]
		}
	}
	close(profile)

	fill_array(_qdisc,"qdisc")
	fill_array(_class,"class")
	fill_array(_filter,"filter")


	###
	#
	# dump the tree starting from a root qdisc
	#
	iterate(_qdisc["root"],1)
	
}
